# -*- coding: utf-8 -*-

"""Top-level package for LyricsMaster."""

__author__ = """SekouD"""
__email__ = 'sekoud.python@gmail.com'
__version__ = '2.7.21'

from .providers import LyricWiki, AzLyrics, Genius
from .utils import TorController



# -*- coding: utf-8 -*-

"""Models the domain objects.

Defines classes for Song, Album and Discography.

"""

import os
from codecs import open
from .utils import set_save_folder, normalize


class Song(object):
    """
    Song class.

    :param title: string.
        Song title.
    :param album: string.
        Album title.
    :param artist: string.
        Author name.
    :param lyrics: string.
        Lyrics of the song.
    :param writers: string.
        List of the song's writers.
    """
    __slots__ = ('title', 'album', 'artist', 'lyrics', 'writers')

    def __init__(self, title, album, artist, lyrics=None, writers=None):
        self.title = title
        self.album = album
        self.artist = artist
        self.lyrics = lyrics
        self.writers = writers

    def __repr__(self):
        return '{0}.{1}({2}, {3}, {4})'.format(__name__, self.__class__.__name__, self.title, self.album, self.artist)

    def save(self, folder=None):
        """
        Saves the lyrics of the song in the supplied folder.
        If no folder is supplied, 'folder' is set to {user}/Documents/lyricsmaster/
        The lyrics of a song are saved in the folder /artist/album/song_title.txt

        :param folder: string.
            path to save folder.
        """
        folder = set_save_folder(folder)
        if self.lyrics:
            artist = normalize(self.artist)
            album = normalize(self.album)
            save_path = os.path.join(folder, artist, album)
            if not os.path.exists(save_path):
                os.makedirs(save_path)
            file_name = normalize(self.title)
            with open(os.path.join(save_path, file_name + ".txt"), "w", encoding="utf-8") as file:
                file.write(self.lyrics)


class Album(object):
    """
    Album Class.
    The Album class follows the Iterable protocol and can be iterated over the songs.

    :param title: string.
        Album title.
    :param artist: string.
        Artist name.
    :param release_date: string.
        Release date.
    :param songs: list.
        List of Songs objects.
    """
    __slots__ = ('__idx__', 'title', 'artist', 'release_date', 'songs')

    def __init__(self, title, artist, songs, release_date='Unknown'):
        self.__idx__ = 0
        self.title = title
        self.artist = artist
        self.release_date = release_date
        self.songs = songs

    def __repr__(self):
        return '{0}.{1}({2}, {3})'.format(__name__, self.__class__.__name__, self.title,self.artist)

    def __len__(self):
        return len(self.songs)

    def __iter__(self):
        return self

    def __next__(self):
        self.__idx__ += 1
        try:
            return self.songs[self.__idx__ - 1]
        except IndexError:
            self.__idx__ = 0
            raise StopIteration

    def __reversed__(self):
        return reversed(self.songs)

    next = __next__ # Python 2.7 compatibility for iterator protocol

    def save(self, folder=None):
        """
        Saves the album in the supplied folder.

        :param folder: string.
            path to save folder.
        """
        for song in self.songs:
            if song:
                song.save(folder)


class Discography(object):
    """
    Discography Class.
    The Discography class follows the Iterable protocol and can be iterated over the albums.

    :param artist: string.
        Artist name.
    :param albums: list.
        List of Album objects.
    """
    __slots__ = ('__idx__', 'artist', 'albums')

    def __init__(self, artist, albums):
        self.__idx__ = 0
        self.artist = artist
        self.albums = albums

    def __repr__(self):
        return '{0}.{1}({2})'.format(__name__, self.__class__.__name__, self.artist)

    def __len__(self):
        return len(self.albums)

    def __iter__(self):
        return self

    def __next__(self):
        self.__idx__ += 1
        try:
            return self.albums[self.__idx__ - 1]
        except IndexError:
            self.__idx__ = 0
            raise StopIteration

    def __reversed__(self):
        return reversed(self.albums)

    next = __next__ # Python 2.7 compatibility for iterator protocol

    def save(self, folder=None):
        """
        Saves Discography in the supplied folder.

        :param folder: string.
            Path to save folder.
        """
        for album in self.albums:
            album.save(folder)



# -*- coding: utf-8 -*-

"""Console script for lyricsmaster."""

import click
import lyricsmaster
from .utils import TorController



@click.command()
@click.argument('artist_name')
@click.option('-p', '--provider', default='LyricWiki', help='Lyrics Provider.', type=click.STRING)
@click.option('-a', '--album', default=None, help='Album.', type=click.STRING)
@click.option('-s', '--song', default=None, help='Song.', type=click.STRING)
@click.option('-f', '--folder', default=None, help='Folder where the lyrics will be saved.', type=click.STRING)
@click.option('--tor', default=None, help='Tor service Ip address.', type=click.STRING)
@click.option('--socksport', default=9050, help='Tor SocksPort.', type=click.INT)
@click.option('--controlport', default=None, help='Tor ControlPort.', type=click.INT)
@click.option('--controlpath', default=None, help='Tor ControlPath.', type=click.STRING)
@click.option('--password', default='', help='Password for Tor ControlPort.', type=click.STRING)
def main(artist_name, provider, album, song, folder, tor, socksport, controlport, controlpath, password):
    """Console script for lyricsmaster."""
    provider = getattr(lyricsmaster, provider)
    if tor:
        if controlport:
            provider_instance = provider(
                TorController(ip=tor, socksport=socksport, controlport=controlport, password=password))
        elif controlpath:
            provider_instance = provider(
                TorController(ip=tor, socksport=socksport, controlport=controlpath, password=password))
        else:
            provider_instance = provider(TorController(ip=tor, socksport=socksport))
    else:
        provider_instance = provider()
    results = provider_instance.get_lyrics(artist_name, album=album, song=song)
    results.save(folder=folder)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Utilities."""

import os
import re
from stem import Signal
from stem.control import Controller
from urllib3.contrib.socks import SOCKSProxyManager
import certifi

import gevent.monkey
import socket

# Python 2.7 compatibility
# Works for Python 2 and 3
try:
    from importlib import reload
except ImportError:
    try:
        from imp import reload
    except:
        pass

# Python 2.7 compatibility
# Works for Python 2 and 3
try:
    basestring
except NameError:
    basestring = str


def normalize(value):
    """

    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.

    :param value: string.
        String.
    :return: string.
        Cleaned string.
    """
    value = re.sub('[^\w\s-]', '', value).strip()
    value = re.sub('[-\s]+', '-', value)
    return value


def set_save_folder(folder):
    """
    Sets the folder in which lyrics will be downloaded and saved.

    :param folder: string.
        Folder path.
    :return: string.
        Folder path.
    """
    if not folder:
        folder = os.path.join(os.path.expanduser("~"), 'Documents', 'LyricsMaster')
    else:
        folder = os.path.join(folder, 'LyricsMaster')
    return folder


class TorController:
    """
    Controller class for Tor client.

    Allows the Api to make requests over the Tor network.
    If 'controlport' is None, the library will use the default timing to renew the Tor circuit.
    If 'controlport' is passed as an argument, the library will create a new Tor circuit for each new album downloaded.
    See https://www.torproject.org/docs/tor-manual.html.en for more information on how Tor works.

    :param ip: string.
        The IP adress of the Tor proxy.
    :param socksport: integer.
        The SOCKSPORT port number for Tor.
    :param controlport: integer or string.
        The CONTROLPORT port number for Tor or the unix path to the CONTROLPATH.
    :param password: string.
        The password or control_auth_cookie to authenticate on the Tor CONTROLPORT.
    """

    def __init__(self, ip='127.0.0.1', socksport=9050, controlport=None, password=''):
        self.ip = ip
        self.socksport = socksport
        self.controlport = controlport
        self.password = password

    def __repr__(self):
        return '{0}.{1}({2}, {3}, {4}, {5})'.format(__name__, self.__class__.__name__, self.ip, self.socksport,
                                                    self.controlport, self.password)

    def get_tor_session(self):
        """
        Configures and create the session to use a Tor Socks proxy.

        :return: urllib3.SOCKSProxyManager object.
        """
        user_agent = {'user-agent':
                'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2228.0 Safari/537.36'}
        session = SOCKSProxyManager('socks5://{0}:{1}'.format(self.ip, self.socksport), cert_reqs='CERT_REQUIRED',
                                    ca_certs=certifi.where(), headers=user_agent)
        return session

    def renew_tor_circuit(self):
        """
        Renews the Tor circuit.
        Sends a NEWNYM message to the Tor network to create a new circuit.

        :return: bool.
            Whether a new tor ciruit was created.

        """

        def renew_circuit(password):
            """
            Sends a NEWNYM message to the Tor network to create a new circuit.

            :param password:
            :return: bool.
            """
            controller.authenticate(password=password)
            if controller.is_newnym_available():  # true if tor would currently accept a NEWNYM signal.
                controller.signal(Signal.NEWNYM)
                print('New Tor circuit created')
                result = True
            else:
                delay = controller.get_newnym_wait()
                print('Delay to create new Tor circuit: {0}s'.format(delay))
                result = False
            return result

        # Needs to reload the default socket to be able to send the is_newnym_avilable and get_newnym_wait signals
        reload(socket)
        if isinstance(self.controlport, int):
            with Controller.from_port(port=self.controlport) as controller:
                is_renewed = renew_circuit(self.password)
        elif isinstance(self.controlport, basestring):
            with Controller.from_socket_file(path=self.controlport) as controller:
                is_renewed = renew_circuit(self.password)
        else:
            is_renewed = False
        gevent.monkey.patch_socket()
        return is_renewed



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `lyricsmaster` package."""

import os
import sys
import codecs

import pytest
from click.testing import CliRunner

from bs4 import BeautifulSoup, Tag

from lyricsmaster import models
from lyricsmaster import cli
from lyricsmaster.providers import LyricWiki, AzLyrics, Genius, Lyrics007, \
    MusixMatch
from lyricsmaster.utils import TorController, normalize

try:
    basestring  # Python 2.7 compatibility
except NameError:
    basestring = str

import gevent.monkey

# Works for Python 2 and 3
try:
    from importlib import reload
except ImportError:
    try:
        from imp import reload
    except:
        pass


def socket_is_patched():
    return gevent.monkey.is_module_patched('socket')


# Boolean tests
python_is_outdated = '2.7' in sys.version or '3.3' in sys.version
is_appveyor = 'APPVEYOR' in os.environ
is_travis = 'TRAVIS' in os.environ

providers = [MusixMatch(), LyricWiki(), Genius(), Lyrics007()]

real_singer = {'name': 'The Notorious B.I.G.', 'album': 'Ready to Die (1994)',
               'songs': [{'song': 'Things Done Changed',
                          'lyrics': 'Remember back in the days...'},
                         {'song': 'Things Done Changed',
                          'lyrics': 'Remember back in the days...'}]
               }
fake_singer = {'name': 'Fake Rapper', 'album': "In my mom's basement",
               'song': 'I fap',
               'lyrics': 'Everyday I fap furiously...'}

provider_strings = {
    'LyricWiki': {'artist_name': 'The_Notorious_B.I.G.',
                  'artist_url': 'http://lyrics.wikia.com/wiki/The_Notorious_B.I.G.',
                  'song_url': 'http://lyrics.wikia.com/wiki/The_Notorious_B.I.G.:Things_Done_Changed',
                  'fake_url': 'http://lyrics.wikia.com/wiki/Things_Done_Changed:Things_Done_Changed_fake_url'},
    'AzLyrics': {'artist_name': 'The Notorious B.I.G.',
                 'artist_url': 'https://www.azlyrics.com/n/notorious.html',
                 'song_url': 'https://www.azlyrics.com/lyrics/notoriousbig/thingsdonechanged.html',
                 'fake_url': 'https://www.azlyrics.com/lyrics/notoriousbig/thingsdonechanged_fake.html'},
    'Genius': {'artist_name': 'The-notorious-big',
               'artist_url': 'https://genius.com/artists/The-notorious-big',
               'song_url': 'https://genius.com/The-notorious-big-things-done-changed-lyrics',
               'fake_url': 'https://genius.com/The-notorious-big-things-done-changed-lyrics_fake'},
    'Lyrics007': {'artist_name': 'The Notorious B.I.G.',
                  'artist_url': 'https://www.lyrics007.com/artist/the-notorious-b-i-g/TVRJMk5EQT0=',
                  'song_url': 'https://www.lyrics007.com/Notorious%20B.i.g.%20Lyrics/Things%20Done%20Changed%20Lyrics.html',
                  'fake_url': 'https://www.lyrics007.com/Notorious%20B.i.g.%20Lyrics/Things%20Done%20Changed%20fake_Lyrics.html'},
    'MusixMatch': {'artist_name': 'The-Notorious-B-I-G',
                   'artist_url': 'https://www.musixmatch.com/artist/The-Notorious-B-I-G',
                   'song_url': 'https://www.musixmatch.com/lyrics/The-Notorious-B-I-G/Things-Done-Changed',
                   'fake_url': 'https://www.musixmatch.com/lyrics/The-Notorious-B-I-G/Things-Done-Changed_fake'},
}

songs = [models.Song(real_singer['songs'][0]['song'], real_singer['album'],
                     real_singer['name'],real_singer['songs'][0]['lyrics']),
         models.Song(real_singer['songs'][1]['song'], real_singer['album'],
                     real_singer['name'], real_singer['songs'][1]['lyrics'])]


class TestSongs:
    """Tests for Song Class."""
    song = songs[0]

    def test_song(self):
        assert self.song.__repr__() == 'lyricsmaster.models.Song({0}, {1}, {2})'.format(
            real_singer['songs'][0]['song'],
            real_singer['album'],
            real_singer['name'])

    def test_song_save(self):
        self.song.save()
        path = os.path.join(os.path.expanduser("~"), 'Documents',
                            'LyricsMaster', normalize(real_singer['name']),
                            normalize(real_singer['album']),
                            'Things-Done-Changed.txt')
        assert os.path.exists(path)
        folder = os.path.join(os.path.expanduser("~"), 'Documents',
                              'test_lyricsmaster_save')
        self.song.save(folder)
        path = os.path.join(folder, 'LyricsMaster',
                            normalize(real_singer['name']),
                            normalize(real_singer['album']),
                            'Things-Done-Changed.txt')
        assert os.path.exists(path)
        with codecs.open(path, 'r', encoding='utf-8') as file:
            assert self.song.lyrics == file.readlines()[0]


class TestAlbums:
    """Tests for Album Class."""

    album = models.Album(real_singer['album'], real_singer['name'], songs, '2017')

    def test_album(self):
        assert self.album.__idx__ == 0
        assert self.album.title == real_singer['album']
        assert self.album.artist == real_singer['name']
        assert self.album.__repr__() == 'lyricsmaster.models.Album({0}, {1})'.format(
            real_singer['album'],
            real_singer['name'])

    def test_album_isiter(self):
        assert len(self.album) == 2
        assert [elmt for elmt in self.album] == songs
        for x, y in zip(reversed(self.album), reversed(self.album.songs)):
            assert x == y

    def test_album_save(self):
        self.album.save()
        for song in self.album.songs:
            artist = normalize(song.artist)
            album = normalize(song.album)
            title = normalize(song.title)
            path = os.path.join(os.path.expanduser("~"), 'Documents',
                                'LyricsMaster', artist, album, title + '.txt')
            assert os.path.exists(path)
            with codecs.open(path, 'r', encoding='utf-8') as file:
                assert song.lyrics == '\n'.join(file.readlines())


class TestDiscography:
    """Tests for Discography Class."""

    albums = [models.Album(real_singer['album'], real_singer['name'], songs, '2017'),
              models.Album(real_singer['album'], real_singer['name'], songs, '2017')]
    discography = models.Discography(real_singer['name'], albums)

    def test_discography(self):
        assert self.discography.__repr__() == 'lyricsmaster.models.Discography({0})'.format(
            real_singer['name'])

    def test_discography_isiter(self):
        assert self.discography.__idx__ == 0
        assert len(self.discography) == 2
        assert [elmt for elmt in self.discography] == self.albums
        for x, y in zip(reversed(self.discography),
                        reversed(self.discography.albums)):
            assert x == y

    def test_discography_save(self):
        self.discography.save()
        for album in self.albums:
            for song in album.songs:
                artist = normalize(song.artist)
                album = normalize(song.album)
                title = normalize(song.title)
                path = os.path.join(os.path.expanduser("~"), 'Documents',
                                    'LyricsMaster', artist, album,
                                    title + '.txt')
                assert os.path.exists(path)
                with codecs.open(path, 'r', encoding='utf-8') as file:
                    assert song.lyrics == '\n'.join(file.readlines())


class TestLyricsProviders:
    """Tests for LyricWiki Class."""

    @pytest.mark.skipif(is_appveyor and '3.3' in sys.version,
                        reason="[SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed (_ssl.c:548) on Appveyor 3.3.")
    @pytest.mark.parametrize('provider', providers)
    def test_get_page(self, provider):
        # Generates unreproducebale errors when when the server does not exist
        # url = 'http://non-existent-url.com'
        # request = provider.get_page(url)
        # assert request is None
        request = provider.get_page('http://www.google.com')
        assert request.status == 200

    @pytest.mark.parametrize('provider', providers)
    def test_clean_string(self, provider):
        assert provider._clean_string(real_singer['name']) == \
               provider_strings[provider.name]['artist_name']

    @pytest.mark.parametrize('provider', providers)
    def test_has_artist(self, provider):
        clean = provider._clean_string
        url = provider._make_artist_url(clean(real_singer['name']))
        page = BeautifulSoup(provider.get_page(url).data, 'lxml')
        assert provider._has_artist(page)
        url = provider._make_artist_url(clean(fake_singer['name']))
        if not url:
            assert url is None
        else:
            page = BeautifulSoup(provider.get_page(url).data, 'lxml')
            assert not provider._has_artist(page)
        pass

    @pytest.mark.parametrize('provider', providers)
    def test_make_artist_url(self, provider):
        clean = provider._clean_string
        assert provider._make_artist_url(clean(real_singer['name'])) == provider_strings[provider.name]['artist_url']

    @pytest.mark.parametrize('provider', providers)
    def test_get_artist_page(self, provider):
        page = provider.get_artist_page(real_singer['name'])
        assert '<!doctype html>' in str(page).lower()
        page = provider.get_artist_page(fake_singer['name'])
        assert page is None

    @pytest.mark.parametrize('provider', providers)
    def test_get_album_page(self, provider):
        if provider.name in ('AzLyrics', 'Genius', 'Lyrics007', 'MusixMatch'):
            return
        else:
            page = provider.get_album_page(real_singer['name'],
                                           fake_singer['album'])
            assert page is None
            page = provider.get_album_page(fake_singer['name'],
                                           fake_singer['album'])
            assert page is None
            page = provider.get_album_page(real_singer['name'],
                                           real_singer['album'])
            assert '<!doctype html>' in str(page).lower()

    @pytest.mark.parametrize('provider', providers)
    def test_has_lyrics(self, provider):
        url = provider_strings[provider.name]['song_url']
        page = BeautifulSoup(provider.get_page(url).data, 'lxml')
        assert provider._has_lyrics(page)
        url = provider_strings[provider.name]['fake_url']
        page = BeautifulSoup(provider.get_page(url).data, 'lxml')
        assert not provider._has_lyrics(page)

    @pytest.mark.parametrize('provider', providers)
    def test_get_lyrics_page(self, provider):
        page = provider.get_lyrics_page(
            provider_strings[provider.name]['song_url'])
        assert '<!doctype html>' in str(page).lower()
        page = provider.get_lyrics_page(
            provider_strings[provider.name]['fake_url'])
        assert page is None

    @pytest.mark.parametrize('provider', providers)
    def test_get_albums(self, provider):
        url = provider_strings[provider.name]['artist_url']
        page = provider.get_page(url).data
        albums = provider.get_albums(page)
        for album in albums:
            assert isinstance(album, Tag)

    @pytest.mark.parametrize('provider', providers)
    def test_get_album_infos(self, provider):
        url = provider_strings[provider.name]['artist_url']
        page = provider.get_page(url).data
        album = provider.get_albums(page)[0]
        album_title, release_date = provider.get_album_infos(album)
        assert isinstance(release_date, basestring)
        # assert album_title.lower() in real_singer['album'].lower() or \
        #        album_title.lower() in 'Demo Tape'.lower() or 'notorious themes' in \
        #        album_title.lower() or 'greatest hits' in album_title.lower()

    @pytest.mark.parametrize('provider', providers)
    def test_extract_lyrics(self, provider):
        page = provider.get_lyrics_page(provider_strings[provider.name]['song_url'])
        lyrics_page = BeautifulSoup(page, 'lxml')
        lyrics = provider.extract_lyrics(lyrics_page)
        assert isinstance(lyrics, basestring)
        assert 'Remember back in the days'.lower() in lyrics.lower()
        assert "Don't ask me why I'm".lower() in lyrics.lower()

    @pytest.mark.parametrize('provider', [prov for prov in providers if
                                          not prov.name in ('Lyrics007',
                                                            'LyricWiki')])
    def test_extract_writers(self, provider):
        page = provider.get_lyrics_page(provider_strings[provider.name]['song_url'])
        lyrics_page = BeautifulSoup(page, 'lxml')
        writers = provider.extract_writers(lyrics_page)
        assert isinstance(writers, basestring)
        assert "c. wallace" in writers.lower() or "notorious" in writers.lower()\
               or "christopher wallace" in writers.lower() or writers == ''

    @pytest.mark.parametrize('provider', providers)
    def test_get_songs(self, provider):
        artist_page = provider.get_artist_page(real_singer['name'])
        album = provider.get_albums(artist_page)[0]
        song_links = provider.get_songs(album)
        for link in song_links:
            assert isinstance(link, Tag)

    @pytest.mark.parametrize('provider', providers)
    def test_create_song(self, provider):
        artist_page = provider.get_artist_page(real_singer['name'])
        album = provider.get_albums(artist_page)[0]
        song_links = provider.get_songs(album)
        song_links[-1].attrs['href'] = provider_strings[provider.name]['fake_url']
        fail_song = provider.create_song(song_links[-1], real_singer['name'], real_singer['album'])
        assert fail_song is None
        good_song = provider.create_song(song_links[0], real_singer['name'], real_singer['album'])
        assert isinstance(good_song, models.Song)
        assert isinstance(good_song.title, basestring)
        assert good_song.album == real_singer['album']
        assert good_song.artist == real_singer['name']
        assert isinstance(good_song.lyrics, basestring)
        # Tests existing song with known empty lyrics
        if provider.name == 'LyricWiki':
            tag = '<a href="http://lyrics.wikia.com/wiki/Reggie_Watts:Feel_The_Same" class="new" title="Reggie Watts:Feel The Same (page does not exist)">Feel the Same</a>'
            page = BeautifulSoup(tag, 'lxml')
            page.attrs[
                'title'] = "Reggie Watts:Feel The Same (page does not exist)"
            page.attrs[
                'href'] = "http://lyrics.wikia.com/wiki/Reggie_Watts:Feel_The_Same"
            non_existent_song = provider.create_song(page, real_singer['name'],
                                                     real_singer['album'])
            assert non_existent_song == None

    @pytest.mark.parametrize('provider', providers)
    def test_get_lyrics(self, provider):
        discography = provider.get_lyrics(fake_singer['name'])
        assert discography is None
        discography = provider.get_lyrics('Reggie Watts', 'Why $#!+ So Crazy?')
        assert isinstance(discography, models.Discography)
        discography = provider.get_lyrics('Reggie Watts', 'Why $#!+ So Crazy?',
                                          'Fuck Shit Stack')
        assert isinstance(discography, models.Discography)


class TestCli:
    """Tests for Command Line Interface."""

    def test_command_line_interface(self):
        artist = 'Reggie Watts'
        runner = CliRunner()
        result = runner.invoke(cli.main, [artist, '-a', 'Why $#!+ So Crazy?', '-s', 'Fuck Shit Stack'])
        assert result.exit_code == 0
        assert 'Why $#!+ So Crazy?' in result.output
        help_result = runner.invoke(cli.main, ['--help'])
        assert help_result.exit_code == 0
        assert 'Show this message and exit.' in help_result.output


# If tests involving Tor are run first, the following tests fail with error: 'an integer is required (got type object)'
class TestTor:
    """Tests for Tor functionality."""
    tor_basic = TorController()
    if is_travis or (is_appveyor and python_is_outdated):
        tor_advanced = TorController(controlport='/var/run/tor/control',
                                     password='password')
    else:
        tor_advanced = TorController(controlport=9051, password='password')

    non_anon_provider = LyricWiki()
    provider = LyricWiki(tor_basic)
    provider2 = LyricWiki(tor_advanced)

    @pytest.mark.skipif(is_appveyor,
                        reason="Tor error on ApppVeyor.")
    def test_anonymisation(self):
        real_ip = self.non_anon_provider.get_page("http://httpbin.org/ip").data
        anonymous_ip = self.provider.get_page("http://httpbin.org/ip").data
        assert real_ip != anonymous_ip

    # this function is tested out in travis using a unix path as a control port instead of port 9051.
    # for now gets permission denied on '/var/run/tor/control' in Travis CI
    @pytest.mark.skipif(is_appveyor or is_travis,
                        reason="Tor error on CI.")
    def test_renew_tor_session(self):
        real_ip = self.non_anon_provider.get_page("http://httpbin.org/ip").data
        anonymous_ip = self.provider2.get_page("http://httpbin.org/ip").data
        assert real_ip != anonymous_ip
        new_tor_circuit = self.provider2.tor_controller.renew_tor_circuit()
        real_ip2 = self.non_anon_provider.get_page("http://httpbin.org/ip").data
        anonymous_ip2 = self.provider2.get_page("http://httpbin.org/ip").data
        assert real_ip2 != anonymous_ip2
        assert new_tor_circuit == True

    @pytest.mark.skipif(is_appveyor,
                        reason="Tor error on ApppVeyor.")
    def test_get_lyrics_tor_basic(self):
        discography = self.provider.get_lyrics(
            'Reggie Watts', 'Why $#!+ So Crazy?',
            'Fuck Shit Stack')  # put another realsinger who has not so many songs to speed up testing.
        assert isinstance(discography, models.Discography)

    @pytest.mark.skipif(is_appveyor or is_travis,
                        reason="Tor error on CI.")
    def test_get_lyrics_tor_advanced(self):
        discography = self.provider2.get_lyrics(
            'Reggie Watts', 'Why $#!+ So Crazy?', 'Fuck Shit Stack')
        assert isinstance(discography, models.Discography)

    @pytest.mark.skipif(is_appveyor,
                        reason="Tor error on ApppVeyor.")
    def test_command_line_interface_tor(self):
        artist = 'Reggie Watts'
        runner = CliRunner()
        result_tor1 = runner.invoke(cli.main,
                                    [artist, '-a', 'Why $#!+ So Crazy?', '-s',
                                     'Fuck Shit Stack', '--tor', '127.0.0.1',
                                     '--controlport', '9051', '--password',
                                     'password'])
        assert result_tor1.exit_code == 0
        assert 'Downloading Why $#!+ So Crazy?' in result_tor1.output

    @pytest.mark.skipif(is_appveyor,
                        reason="Tor error on ApppVeyor.")
    def test_command_line_interface_tor(self):
        artist = 'Reggie Watts'
        runner = CliRunner()
        result_tor = runner.invoke(cli.main, [artist, '-a', 'Why $#!+ So Crazy?', '-s',
                                              'Fuck Shit Stack', '--tor', '127.0.0.1'])
        assert result_tor.exit_code == 0
        assert 'Downloading Why $#!+ So Crazy?' in result_tor.output



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('docs/usage.rst') as usage_file:
    usage = usage_file.read()

requirements = [
    'Click>=6.0',
    'lxml',
    'beautifulsoup4',
    'urllib3',
    'urllib3[secure]',
    'pysocks',
    'gevent',
    'stem',
    'certifi',
]

setup_requirements = [
    'pytest'
    # TODO(SekouD): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    'Sphinx',
    'docutils',
    'pytest',
    'pytest-cov',
    'Click>=6.0',
    'lxml',
    'beautifulsoup4',
    'requests',
    'requests[socks]',
    'urllib3'
    'PySocks',
    'gevent',
    'stem'
]

setup(
    name='lyricsmaster',
    version='2.7.21',
    description="LyricsMaster is a library for downloading lyrics from multiple lyrics providers. LyricWiki, AzLyrics, Genius, Lyrics007, MusixMatch and other lyrics provider are available",
    long_description=readme + '\n\n' + usage + '\n\n' + history,
    author="SekouD",
    author_email='sekoud.python@gmail.com',
    url='https://github.com/SekouD/lyricsmaster',
    packages=find_packages(include=['lyricsmaster']),
    entry_points={
        'console_scripts': [
            'lyricsmaster=lyricsmaster.cli:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='lyricsmaster lyrics LyricWiki Lyrics Wikia Lyrics007 AzLyrics Genius MusixMatch Tor',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Communications',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



# -*- coding: utf-8 -*-

"""Main module.

This module defines the Api interface for the various Lyrics providers.
All lyrics providers inherit from the base class LyricsProvider.

"""

# We use abstract methods to ensure that all future classes inheriting from LyricsProvider will
# implement the required methods in order to have a nice and consistent API.
from abc import ABCMeta, abstractmethod

import re
import urllib3
import certifi
from bs4 import BeautifulSoup

# We use gevent in order to make asynchronous http requests while downloading lyrics.
# It is also used to patch the socket module to use SOCKS5 instead to interface with the Tor controller.
import gevent.monkey
from gevent.pool import Pool

# Python 2.7 compatibility
# Works for Python 2 and 3
try:
    from importlib import reload
except ImportError:
    try:
        from imp import reload
    except:
        pass

# Importing the app models and utilities
from .models import Song, Album, Discography
from .utils import normalize

class LyricsProvider:
    """
    This is the base class for all Lyrics Providers. If you wish to subclass this class, you must implement all
    the methods defined in this class to be compatible with the LyricsMaster API.
    Requests to fetch songs are executed asynchronously for better performance.
    Tor anonymisation is provided if tor is installed on the system and a TorController is passed at instance creation.

    :param tor_controller: TorController Object.

    """
    __metaclass__ = ABCMeta
    name = ''

    def __init__(self, tor_controller=None):
        if not self.__socket_is_patched():
            gevent.monkey.patch_socket()
        self.tor_controller = tor_controller
        if not self.tor_controller:
            user_agent = {'user-agent': 'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2228.0 Safari/537.36'}
            self.session = urllib3.PoolManager(maxsize=10, cert_reqs='CERT_REQUIRED', ca_certs=certifi.where(),
                                               headers=user_agent)
        else:
            self.session = self.tor_controller.get_tor_session()
        self.__tor_status__()

    def __repr__(self):
        return '{0}.{1}({2})'.format(__name__, self.__class__.__name__, self.tor_controller.__repr__())

    def __tor_status__(self):
        """
        Informs the user of the Tor status.

        """
        if not self.tor_controller:
            print('Anonymous requests disabled. The connexion will not be anonymous.')
        elif self.tor_controller and not self.tor_controller.controlport:
            print('Anonymous requests enabled. The Tor circuit will change according to the Tor network defaults.')
        else:
            print('Anonymous requests enabled. The Tor circuit will change for each album.')

    def __socket_is_patched(self):
        """
        Checks if the socket is patched or not.

        :return: bool.
        """
        return gevent.monkey.is_module_patched('socket')

    @abstractmethod
    def _has_lyrics(self, page):
        """
        Must be implemented by children classes conforming to the LyricsMaster API.

        Checks if the lyrics provider has the lyrics for the song or not.

        :param page:
        :return: bool.
        """
        pass

    @abstractmethod
    def _has_artist(self, page):
        """
        Must be implemented by children classes conforming to the LyricsMaster API.

        Check if the artist is in the lyrics provider's database.

        :param page:
        :return: bool.
        """
        pass

    @abstractmethod
    def _make_artist_url(self, artist):
        """
        Must be implemented by children classes conforming to the LyricsMaster API.

        Builds an url for the artist page of the lyrics provider.

        :param artist:
        :return: string or None.
        """
        pass

    @abstractmethod
    def _clean_string(self, text):
        """
        Must be implemented by children classes conforming to the LyricsMaster API.

        Formats the text to conform to the lyrics provider formatting.

        :param text:
        :return: string or None.
        """
        pass

    def get_page(self, url):
        """
        Fetches the supplied url and returns a request object.

        :param url: string.
        :return: urllib3.response.HTTPResponse Object or None.
        """
        if not self.__socket_is_patched():
            gevent.monkey.patch_socket()
        try:
            req = self.session.request('GET', url)
        except Exception as e:
            print(e)
            req = None
            print('Unable to download url ' + url)
        return req

    def get_lyrics(self, artist, album=None, song=None):
        """
        This is the main method of this class.
        Connects to the Lyrics Provider and downloads lyrics for all the albums of the supplied artist and songs.
        Returns a Discography Object or None if the artist was not found on the Lyrics Provider.

        :param artist: string
            Artist name.
        :return: models.Discography object or None.
        """

        raw_html = self.get_artist_page(artist)
        if not raw_html:
            print('{0} was not found on {1}'.format(artist, self.name))
            return None
        albums = self.get_albums(raw_html)
        if album:
            # If user supplied a specific album
            albums = [elmt for elmt in albums if album.lower() in self.get_album_infos(elmt)[0].lower()]
        album_objects = []
        for elmt in albums:
            try:
                album_title, release_date = self.get_album_infos(elmt)
            except ValueError as e:
                pass
                print('Error {0} while downloading {1}'.format(e, album_title))
                continue
            song_links = self.get_songs(elmt)
            if song:
                # If user supplied a specific song
                song_links = [link for link in song_links if song.lower() in link.text.lower()]
            if self.tor_controller and self.tor_controller.controlport:
                # Renew Tor circuit before starting downloads.
                self.tor_controller.renew_tor_circuit()
                self.session = self.tor_controller.get_tor_session()
            print('Downloading {0}'.format(album_title))
            pool = Pool(25)  # Sets the worker pool for async requests. 25 is a nice value to not annoy site owners ;)
            results = [pool.spawn(self.create_song, *(link, artist, album_title)) for link in song_links]
            pool.join()  # Gathers results from the pool
            songs = [song.value for song in results]
            album_obj = Album(album_title, artist, songs, release_date)
            album_objects.append(album_obj)
            print('{0} succesfully downloaded'.format(album_title))
        discography = Discography(artist, album_objects)
        return discography

    def get_artist_page(self, artist):
        """
        Fetches the web page for the supplied artist.

        :param artist: string.
            Artist name.
        :return: string or None.
            Artist's raw html page. None if the artist page was not found.
        """
        artist = self._clean_string(artist)
        url = self._make_artist_url(artist)
        if not url:
            return None
        raw_html = self.get_page(url).data
        artist_page = BeautifulSoup(raw_html, 'lxml')
        if not self._has_artist(artist_page):
            return None
        return raw_html

    def get_lyrics_page(self, url):
        """
        Fetches the web page containing the lyrics at the supplied url.

        :param url: string.
            Lyrics url.
        :return: string or None.
            Lyrics's raw html page. None if the lyrics page was not found.
        """
        raw_html = self.get_page(url).data
        lyrics_page = BeautifulSoup(raw_html, 'lxml')
        if not self._has_lyrics(lyrics_page):
            return None
        return raw_html

    @abstractmethod
    def get_albums(self, raw_artist_page):
        """
        Must be implemented by children classes conforming to the LyricsMaster API.

        Fetches the albums section in the supplied html page.

        :param raw_artist_page: Artist's raw html page.
        :return: list.
            List of BeautifulSoup objects.
        """
        pass

    @abstractmethod
    def get_album_infos(self, tag):
        """
        Must be implemented by children classes conforming to the LyricsMaster API.

        Extracts the Album informations from the tag

        :param tag: BeautifulSoup object.
        :return: tuple(string, string).
            Album title and release date.
        """
        pass

    @abstractmethod
    def get_songs(self, album):
        """
        Must be implemented by children classes conforming to the LyricsMaster API.

        Fetches the links to the songs of the supplied album.

        :param album: BeautifulSoup object.
        :return: List of BeautifulSoup Link objects.
        """
        pass

    @abstractmethod
    def create_song(self, link, artist, album_title):
        """
        Must be implemented by children classes conforming to the LyricsMaster API.

        Creates a Song object.

        :param link: BeautifulSoup Link object.
        :param artist: string.
        :param album_title: string.
        :return: models.Song object or None.
        """
        pass

    @abstractmethod
    def extract_lyrics(self, song):
        """
        Must be implemented by children classes conforming to the LyricsMaster API.

        Extracts the lyrics from the lyrics page of the supplied song.

        :param lyrics_page: BeautifulSoup Object.
            BeautifulSoup lyrics page.
        :return: string or None.
            Formatted lyrics.
        """
        pass

    @abstractmethod
    def extract_writers(self, lyrics_page):
        """
        Must be implemented by children classes conforming to the LyricsMaster API.

        Extracts the writers from the lyrics page of the supplied song.

        :param lyrics_page: BeautifulSoup Object.
            BeautifulSoup lyrics page.
        :return: string or None.
            Song writers.
        """
        pass


class LyricWiki(LyricsProvider):
    """
    Class interfacing with http://lyrics.wikia.com .
    This class is used to retrieve lyrics from LyricWiki.

    """
    base_url = 'http://lyrics.wikia.com'
    name = 'LyricWiki'

    def _has_lyrics(self, lyrics_page):
        """
        Checks if the lyrics provider has the lyrics for the song or not.

        :param lyrics_page:
        :return: bool.
        """
        return not lyrics_page.find("div", {'class': 'noarticletext'})

    _has_artist = _has_lyrics

    def _make_artist_url(self, artist):
        """
        Builds an url for the artist page of the lyrics provider.

        :param artist:
        :return: string.
        """
        url = self.base_url + '/wiki/' + artist
        return url

    def get_album_page(self, artist, album):
        """
        Fetches the album page for the supplied artist and album.

        :param artist: string.
            Artist name.
        :param album: string.
            Album title.
        :return: string or None.
            Album's raw html page. None if the album page was not found.
        """
        artist = self._clean_string(artist)
        album = self._clean_string(album)
        url = self.base_url + '/wiki/' + artist + ':' + album
        raw_html = self.get_page(url).data
        album_page = BeautifulSoup(raw_html, 'lxml')
        if album_page.find("div", {'class': 'noarticletext'}):
            return None
        return raw_html

    def get_albums(self, raw_artist_page):
        """
        Fetches the albums section in the supplied html page.

        :param raw_artist_page: Artist's raw html page.
        :return: list.
            List of BeautifulSoup objects.
        """
        artist_page = BeautifulSoup(raw_artist_page, 'lxml')
        albums = [tag for tag in artist_page.find_all("span", {'class': 'mw-headline'}) if
                  tag.attrs['id'] not in ('Additional_information', 'External_links')]
        return albums

    def get_album_infos(self, tag):
        """
        Extracts the Album informations from the tag

        :param tag: BeautifulSoup object.
        :return: tuple(string, string).
            Album title and release date.
        """
        try:
            i = tag.text.index(' (')
            release_date = re.findall(r'\(([^()]+)\)', tag.text)[0]
        except ValueError:
            i = -1
            release_date = 'Unknown'
        album_title = tag.text[:i]

        return album_title, release_date

    def get_songs(self, album):
        """
        Fetches the links to the songs of the supplied album.

        :param album: BeautifulSoup object.
        :return: List of BeautifulSoup Link objects.
        """
        parent_node = album.parent
        while parent_node.name != 'ol':
            parent_node = parent_node.next_sibling
        song_links = [elmt.find('a') for elmt in parent_node.find_all('li')]
        return song_links

    def create_song(self, link, artist, album_title):
        """
        Creates a Song object.

        :param link: BeautifulSoup Link object.
        :param artist: string.
        :param album_title: string.
        :return: models.Song object or None.
        """
        if not link.attrs['href'].startswith(self.base_url):
            song_url = self.base_url + link.attrs['href']
        else:
            song_url = link.attrs['href']
        song_title = link.attrs['title']
        song_title = song_title[song_title.index(':') + 1:]
        if '(page does not exist' in song_title:
            return None
        raw_lyrics_page = self.get_lyrics_page(song_url)
        if not raw_lyrics_page:
            return None
        lyrics_page = BeautifulSoup(raw_lyrics_page, 'lxml')
        lyrics = self.extract_lyrics(lyrics_page)
        writers = self.extract_writers(lyrics_page)
        song = Song(song_title, album_title, artist, lyrics, writers)
        return song

    def extract_lyrics(self, lyrics_page):
        """
        Extracts the lyrics from the lyrics page of the supplied song.

        :param lyrics_page: BeautifulSoup Object.
            BeautifulSoup lyrics page.
        :return: string or None.
            Formatted lyrics.
        """
        lyric_box = lyrics_page.find("div", {'class': 'lyricbox'})
        lyrics = '\n'.join(lyric_box.strings)
        return lyrics

    def extract_writers(self, lyrics_page):
        """
        Extracts the writers from the lyrics page of the supplied song.

        :param lyrics_page: BeautifulSoup Object.
            BeautifulSoup lyrics page.
        :return: string or None.
            Song writers.
        """
        writers_box = lyrics_page.find("table", {'class': 'song-credit-box'})
        if writers_box:
            writers = writers_box.find_all('p')[-1].text.strip()
        else:
            writers = None
        return writers

    def _clean_string(self, text):
        """
        Cleans the supplied string and formats it to use in a url.

        :param text: string.
            Text to be cleaned.
        :return: string.
            Cleaned text.
        """
        for elmt in [('#', 'Number_'), ('[', '('), (']', ')'), ('{', '('), ('}', ')'), (' ', '_')]:
            text = text.replace(*elmt)
        return text


class AzLyrics(LyricsProvider):
    """
    Class interfacing with https://azlyrics.com .
    This class is used to retrieve lyrics from AzLyrics.

    """
    base_url = 'https://www.azlyrics.com'
    search_url = 'https://search.azlyrics.com/search.php?q='
    name = 'AzLyrics'

    def _has_lyrics(self, lyrics_page):
        """
        Checks if the lyrics provider has the lyrics for the song or not.

        :param lyrics_page:
        :return: bool.
        """
        if lyrics_page.find("div", {'class': 'lyricsh'}):
            return True
        else:
            return False

    def _has_artist(self, page):
        """
        Check if the artist is in the lyrics provider's database.

        :param lyrics_page:
        :return: bool.
        """
        if page.find("div", {'id': 'listAlbum'}):
            return True
        else:
            return False

    def _has_artist_result(self, page):
        """
        Checks if the lyrics provider has the lyrics for the song or not.

        :param lyrics_page:
        :return: bool.
        """
        artist_result = page.find("div", {'class': 'panel-heading'})
        if artist_result.find('b').text == 'Artist results:':
            return True
        else:
            return False

    def _make_artist_url(self, artist):
        """
        Builds an url for the artist page of the lyrics provider.

        :param artist:
        :return: string.
        """
        return self.search(artist)

    def search(self, artist):
        """
        Searches for the artist in the supplier's database.

        :param artist: Artist's name.
        :return: url or None.
            Url to the artist's page if found. None if not Found.
        """
        artist = artist.replace(' ', '+')
        if artist.lower().startswith('the'):
            artist = artist[4:]
        url = self.search_url + artist
        search_results = self.get_page(url).data
        results_page = BeautifulSoup(search_results, 'lxml')
        if not self._has_artist_result(results_page):
            return None
        target_node = results_page.find("div", {'class': 'panel-heading'}).find_next_sibling("table")
        artist_url = target_node.find('a').attrs['href']
        if not artist_url:
            return None
        if not artist_url.startswith(self.base_url):
            artist_url = self.base_url + artist_url
        return artist_url

    def get_albums(self, raw_artist_page):
        """
        Fetches the albums section in the supplied html page.

        :param raw_artist_page: Artist's raw html page.
        :return: list.
            List of BeautifulSoup objects.
        """
        artist_page = BeautifulSoup(raw_artist_page, 'lxml')
        albums = [tag for tag in artist_page.find_all("div", {'id': 'listAlbum'})]
        return albums

    def get_album_infos(self, tag):
        """
        Extracts the Album informations from the tag

        :param tag: BeautifulSoup object.
        :return: tuple(string, string).
            Album title and release date.
        """
        album_infos = tag.find("div", {'class': 'album'}).text
        album_title = re.findall(r'"([^"]*)"', album_infos)[0]
        try:
            release_date = re.findall(r'\(([^()]+)\)', tag.text)[0]
        except ValueError:
            release_date = 'Unknown'
        return album_title, release_date

    def get_songs(self, album):
        """
        Fetches the links to the songs of the supplied album.

        :param album: BeautifulSoup object.
        :return: List of BeautifulSoup Link objects.
        """
        song_links = album.find_all('a')
        song_links = [song for song in song_links if 'href' in song.attrs]
        return song_links

    def create_song(self, link, artist, album_title):
        """
        Creates a Song object.

        :param link: BeautifulSoup Link object.
        :param artist: string.
        :param album_title: string.
        :return: models.Song object or None.
        """
        song_title = link.text
        raw_lyrics_page = self.get_lyrics_page(self.base_url + link.attrs['href'].replace('..', ''))
        if not raw_lyrics_page:
            return None
        lyrics_page = BeautifulSoup(raw_lyrics_page, 'lxml')
        lyrics = self.extract_lyrics(lyrics_page)
        writers = self.extract_writers(lyrics_page)
        song = Song(song_title, album_title, artist, lyrics, writers)
        return song

    def extract_lyrics(self, lyrics_page):
        """
        Extracts the lyrics from the lyrics page of the supplied song.

        :param lyrics_page: BeautifulSoup Object.
            BeautifulSoup lyrics page.
        :return: string.
            Formatted lyrics.
        """
        lyric_box = lyrics_page.find("div", {"class": None, "id": None})
        lyrics = ''.join(lyric_box.strings)
        return lyrics

    def extract_writers(self, lyrics_page):
        """
        Extracts the writers from the lyrics page of the supplied song.

        :param lyrics_page: BeautifulSoup Object.
            BeautifulSoup lyrics page.
        :return: string or None.
            Song writers or None.
        """
        writers_box = lyrics_page.find("div", {'class': 'smt'})
        if writers_box:
            writers = writers_box.text.strip()
        else:
            writers = None
        return writers

    def _clean_string(self, text):
        """
        Cleans the supplied string and formats it to use in a url.

        :param text: string.
            Text to be cleaned.
        :return: string.
            Cleaned text.
        """
        return text


class Genius(LyricsProvider):
    """
    Class interfacing with https://genius.com .
    This class is used to retrieve lyrics from Genius.

    """
    base_url = 'https://genius.com'
    search_url = base_url + '/search?q='
    name = 'Genius'

    def _has_lyrics(self, page):
        """
        Checks if the lyrics provider has the lyrics for the song or not.

        :param lyrics_page:
        :return: bool.
        """
        if page.find("div", {'class': 'song_body-lyrics'}):
            return True
        else:
            return False

    def _has_artist(self, page):
        """
        Check if the artist is in the lyrics provider's database.

        :param lyrics_page:
        :return: bool.
        """
        if not page.find("div", {'class': 'render_404'}):
            return True
        else:
            return False

    def _make_artist_url(self, artist):
        """
        Builds an url for the artist page of the lyrics provider.

        :param artist:
        :return: string.
        """
        url = self.base_url + '/artists/' + artist
        return url

    def get_albums(self, raw_artist_page):
        """
        Fetches the albums section in the supplied html page.

        :param raw_artist_page: Artist's raw html page.
        :return: list.
            List of BeautifulSoup objects.
        """
        artist_page = BeautifulSoup(raw_artist_page, 'lxml')
        albums_link = artist_page.find("a", {'class': 'full_width_button'})
        albums_link = albums_link.attrs['href'].replace('songs?', 'albums?')
        albums_page = BeautifulSoup(self.get_page(self.base_url + albums_link).data, 'lxml')
        albums = [tag for tag in albums_page.find_all("a", {'class': 'album_link'})]
        return albums

    def get_album_infos(self, tag):
        """
        Extracts the Album informations from the tag

        :param tag: BeautifulSoup object.
        :return: tuple(string, string).
            Album title and release date.
        """
        album_title = tag.text
        album_page = BeautifulSoup(self.get_page(self.base_url + tag.attrs['href']).data, 'lxml')
        info_box = album_page.find("div", {'class': 'header_with_cover_art-primary_info'})
        metadata = [elmt for elmt in info_box.find_all("div", {'class': 'metadata_unit'}) if elmt.text.startswith('Released')]
        try:
            release_date = metadata[0].text
        except IndexError:
            release_date = 'Unknown'
        except ValueError:
            release_date = 'Unknown'
        return album_title, release_date

    def get_songs(self, album):
        """
        Fetches the links to the songs of the supplied album.

        :param album: BeautifulSoup object.
        :return: List of BeautifulSoup Link objects.
        """
        album_page = BeautifulSoup(self.get_page(self.base_url + album.attrs['href']).data, 'lxml')
        song_links = album_page.find_all("div", {'class': 'chart_row chart_row--light_border chart_row--full_bleed_left chart_row--align_baseline chart_row--no_hover'})
        song_links = [song.find('a') for song in song_links]
        return song_links

    def create_song(self, link, artist, album_title):
        """
        Creates a Song object.

        :param link: BeautifulSoup Link object.
        :param artist: string.
        :param album_title: string.
        :return: models.Song object or None.
        """
        if not link.attrs['href'].startswith(self.base_url):
            song_url = self.base_url + link.attrs['href']
        else:
            song_url = link.attrs['href']
        song_title = link.text.strip('\n').split('\n')[0].lstrip()
        raw_lyrics_page = self.get_lyrics_page(song_url)
        if not raw_lyrics_page:
            return None
        lyrics_page = BeautifulSoup(raw_lyrics_page, 'lxml')
        lyrics = self.extract_lyrics(lyrics_page)
        writers = self.extract_writers(lyrics_page)
        song = Song(song_title, album_title, artist, lyrics, writers)
        return song

    def extract_lyrics(self, lyrics_page):
        """
        Extracts the lyrics from the lyrics page of the supplied song.

        :param lyrics_page: BeautifulSoup Object.
            BeautifulSoup lyrics page.
        :return: string.
            Formatted lyrics.
        """
        lyric_box = lyrics_page.find("div", {"class": 'lyrics'})
        lyrics = ''.join(lyric_box.strings)
        return lyrics

    def extract_writers(self, lyrics_page):
        """
        Extracts the writers from the lyrics page of the supplied song.

        :param lyrics_page: BeautifulSoup Object.
            BeautifulSoup lyrics page.
        :return: string.
            Song writers or None.
        """
        writers_box = [elmt for elmt in lyrics_page.find_all("span", {'class': 'metadata_unit-label'}) if elmt.text == "Written By"]
        if writers_box:
            target_node = writers_box[0].find_next_sibling("span", {'class': 'metadata_unit-info'})
            writers = target_node.text.strip()
        else:
            writers = None
        return writers

    def _clean_string(self, text):
        """
        Cleans the supplied string and formats it to use in a url.

        :param text: string.
            Text to be cleaned.
        :return: string.
            Cleaned text.
        """
        text = normalize(text).lower().capitalize()
        return text


class Lyrics007(LyricsProvider):
    """
    Class interfacing with https://www.lyrics007.com .
    This class is used to retrieve lyrics from Lyrics007.

    """
    base_url = 'https://www.lyrics007.com'
    search_url = base_url + '/search.php?category=artist&q='
    name = 'Lyrics007'

    def _has_lyrics(self, page):
        """
        Checks if the lyrics provider has the lyrics for the song or not.

        :param lyrics_page:
        :return: bool.
        """
        if page.find("div", {'class': 'lyrics'}):
            return True
        else:
            return False

    def _has_artist(self, page):
        """
        Check if the artist is in the lyrics provider's database.

        :param lyrics_page:
        :return: bool.
        """
        if page.find("ul", {'class': 'song_title'}):
            return True
        else:
            return False

    def _has_artist_result(self, page):
        """
        Check if the artist is in the lyrics provider's database.

        :param lyrics_page:
        :return: bool.
        """
        artist_link = page.find("div", {'id': 'search_result'}).find('a')
        if artist_link:
            return True
        else:
            return False

    def _make_artist_url(self, artist):
        """
        Builds an url for the artist page of the lyrics provider.

        :param artist:
        :return: string.
        """
        return self.search(artist)

    def search(self, artist):
        """
        Searches for the artist in the supplier's database.

        :param raw_artist_page: Artist's raw html page.
        :return: string or None.
            Artist's url page.
        """
        artist = "".join([c if (c.isalnum() or c == '.') else "+" for c in artist])
        url = self.search_url + artist
        search_results = self.get_page(url).data
        results_page = BeautifulSoup(search_results, 'lxml')
        if not self._has_artist_result(results_page):
            return None
        artist_url = results_page.find("div", {'id': 'search_result'}).find('a').attrs['href']
        if not artist_url:
            return None
        if not artist_url.startswith(self.base_url):
            artist_url = self.base_url + artist_url
        return artist_url

    def get_albums(self, raw_artist_page):
        """
        Fetches the albums section in the supplied html page.

        :param raw_artist_page: Artist's raw html page.
        :return: list.
            List of BeautifulSoup objects.
        """
        artist_page = BeautifulSoup(raw_artist_page, 'lxml')
        albums = [tag for tag in artist_page.find_all('li') if tag.find('b')]
        return albums

    def get_album_infos(self, tag):
        """
        Extracts the Album informations from the tag

        :param tag: BeautifulSoup object.
        :return: tuple(string, string).
            Album title and release date.
        """
        # TODO: Verify order of release_date, album_title = tag.text.split(': ')
        release_date, album_title = tag.text.split(': ')
        return album_title, release_date

    def get_songs(self, album):
        """
        Fetches the links to the songs of the supplied album.

        :param album: BeautifulSoup object.
        :return: List of BeautifulSoup Link objects.
        """
        target_node = album.find_next_sibling("ul")
        song_links = [elmt.find('a') for elmt in target_node.find_all('li') if elmt.find('a')]
        return song_links

    def create_song(self, link, artist, album_title):
        """
        Creates a Song object.

        :param link: BeautifulSoup Link object.
        :param artist: string.
        :param album_title: string.
        :return: models.Song object or None.
        """
        if not link.attrs['href'].startswith(self.base_url):
            song_url = self.base_url + link.attrs['href']
        else:
            song_url = link.attrs['href']
        song_title = link.text
        raw_lyrics_page = self.get_lyrics_page(song_url)
        if not raw_lyrics_page:
            return None
        lyrics_page = BeautifulSoup(raw_lyrics_page, 'lxml')
        lyrics = self.extract_lyrics(lyrics_page)
        writers = self.extract_writers(lyrics_page)
        song = Song(song_title, album_title, artist, lyrics, writers)
        return song

    def extract_lyrics(self, lyrics_page):
        """
        Extracts the lyrics from the lyrics page of the supplied song.

        :param lyrics_page: BeautifulSoup Object.
            BeautifulSoup lyrics page.
        :return: string.
            Formatted lyrics.
        """
        lyric_box = lyrics_page.find("div", {'class': 'lyrics'})
        lyrics = '\n'.join(lyric_box.strings)
        return lyrics

    def extract_writers(self, lyrics_page):
        """
        Extracts the writers from the lyrics page of the supplied song.

        :param lyrics_page: BeautifulSoup Object.
            BeautifulSoup lyrics page.
        :return: string.
            Song writers or None.
        """
        writers_box = [elmt for elmt in lyrics_page.strings if
                       elmt.lower().startswith('writers:') or elmt.lower().startswith('writer:')]
        if writers_box:
            writers = writers_box[0].strip()
        else:
            writers = None
        return writers

    def _clean_string(self, text):
        """
        Cleans the supplied string and formats it to use in a url.

        :param text: string.
            Text to be cleaned.
        :return: string.
            Cleaned text.
        """
        return text


class MusixMatch(LyricsProvider):
    """
    Class interfacing with https://www.musixmatch.com .
    This class is used to retrieve lyrics from MusixMatch.

    """
    base_url = 'https://www.musixmatch.com'
    search_url = base_url + '/search/{0}/artists'
    name = 'MusixMatch'

    def _has_lyrics(self, page):
        """
        Checks if the lyrics provider has the lyrics for the song or not.

        :param lyrics_page:
        :return: bool.
        """
        if page.find("p", {'class': 'mxm-lyrics__content '}):
            return True
        else:
            return False

    def _has_artist(self, page):
        """
        Check if the artist is in the lyrics provider's database.

        :param lyrics_page:
        :return: bool.
        """
        if page.find("div", {'class': 'artist-page main-wrapper'}):
            return True
        else:
            return False

    def _make_artist_url(self, artist):
        """
        Builds an url for the artist page of the lyrics provider.

        :param artist:
        :return: string.
        """
        return self.base_url + '/artist/' + artist

    def get_albums(self, raw_artist_page):
        """
        Fetches the albums section in the supplied html page.

        :param raw_artist_page: Artist's raw html page.
        :return: list.
            List of BeautifulSoup objects.
        """
        artist_page = BeautifulSoup(raw_artist_page, 'lxml')
        albums_link = artist_page.find("li", {'id': 'albums'})
        albums_link = albums_link.find('a').attrs['href']
        albums_page = BeautifulSoup(self.get_page(self.base_url + albums_link).data, 'lxml')
        albums = [tag for tag in albums_page.find_all("div", {'class': 'media-card-text'})]
        return albums

    def get_album_infos(self, tag):
        """
        Extracts the Album informations from the tag

        :param tag: BeautifulSoup object.
        :return: tuple(string, string).
            Album title and release date.
        """
        album_title = tag.find('h2').text
        try:
            release_date = tag.find('h3').text
        except AttributeError:
            release_date = 'Unknown'
        return album_title, release_date

    def get_songs(self, album):
        """
        Fetches the links to the songs of the supplied album.

        :param album: BeautifulSoup object.
        :return: List of BeautifulSoup Link objects.
        """
        album_page = BeautifulSoup(self.get_page(self.base_url + album.find('a').attrs['href']).data, 'lxml')
        album_div = album_page.find("div", {'class': 'mxm-album__tracks mxm-collection-container'})
        song_links = album_div.find_all("li", {'class': re.compile("^mui-collection__item")})
        song_links = [song.find('a') for song in song_links]
        return song_links

    def create_song(self, link, artist, album_title):
        """
        Creates a Song object.

        :param link: BeautifulSoup Link object.
        :param artist: string.
        :param album_title: string.
        :return: models.Song object or None.
        """
        if not link.attrs['href'].startswith(self.base_url):
            song_url = self.base_url + link.attrs['href']
        else:
            song_url = link.attrs['href']
        song_title = link.text
        raw_lyrics_page = self.get_lyrics_page(song_url)
        if not raw_lyrics_page:
            return None
        lyrics_page = BeautifulSoup(raw_lyrics_page, 'lxml')
        lyrics = self.extract_lyrics(lyrics_page)
        writers = self.extract_writers(lyrics_page)
        song = Song(song_title, album_title, artist, lyrics, writers)
        return song

    def extract_lyrics(self, lyrics_page):
        """
        Extracts the lyrics from the lyrics page of the supplied song.

        :param lyrics_page: BeautifulSoup Object.
            BeautifulSoup lyrics page.
        :return: string.
            Formatted lyrics.
        """
        lyric_box = lyrics_page.find("p", {'class': re.compile("^mxm-lyrics__content")})
        lyrics = '\n'.join(lyric_box.strings)
        return lyrics

    def extract_writers(self, lyrics_page):
        """
        Extracts the writers from the lyrics page of the supplied song.

        :param lyrics_page: BeautifulSoup Object.
            BeautifulSoup lyrics page.
        :return: string.
            Song writers or None.
        """
        writers_box = lyrics_page.find("p", {'class': re.compile("^mxm-lyrics__copyright")})
        if writers_box:
            writers = writers_box.text.strip()
        else:
            writers = None
        return writers

    def _clean_string(self, text):
        """
        Cleans the supplied string and formats it to use in a url.

        :param text: string.
            Text to be cleaned.
        :return: string.
            Cleaned text.
        """
        text = text.replace(' ', '-').replace('.', '-')
        if text[-1] == '-':
            text = text[:-1]
        return text

if __name__ == "__main__":
    pass
